from __future__ import annotations
from typing import Generic, TypeVar, Protocol
import typing

I = TypeVar("I")

# Note: using typing.Iterator as the result creates an indirection that avoids an infinite
# loop when constructing checkers.
class Iterator(Generic[I]):
    def __next__(self) -> I:
        pass
    def __iter__(self) -> typing.Iterator[I]:
        pass

class Iterable(Generic[I]):
    def __iter__(self) -> typing.Iterator[I]:
        pass

class OnlyIterable(Generic[I], Protocol):
    # The __protocol_only__ flag signals that an object wrapped with the protocol should
    # only provide the methods of the protocol.
    __protocol_only__ = True

    def __iter__(self) -> typing.Iterator[I]:
        pass



import sys
import os

def _getEnv(name, conv, default):
    s = os.getenv(name)
    if s is None:
        return default
    try:
        return conv(s)
    except:
        return default

_DEBUG = _getEnv("WYPP_DEBUG", bool, False)

def enableDebug(debug: bool):
    global _DEBUG
    _DEBUG = debug

def isDebug():
    return _DEBUG

def debug(s):
    if _DEBUG:
        sys.stderr.write('[DEBUG] ' + s + '\n')



from writeYourProgram import *
import unittest
from sample import *

setDieOnCheckFailures(True)

class TestSample(unittest.TestCase):
    def test_sample(self):
        check(pointInShape(p2, c1), True)
        check(pointInShape(p3, c2), True)
        check(pointInShape(Point(51, 50), c1), False)
        check(pointInShape(Point(11, 21), s1), True)
        check(pointInShape(Point(49, 59), s1), True)
        check(pointInShape(Point(9, 21), s1), False)
        check(pointInShape(Point(11, 19), s1), False)
        check(pointInShape(Point(51, 59), s1), False)
        check(pointInShape(Point(49, 61), s1), False)
        check(pointInShape(Point(40, 30), c2), True)
        check(pointInShape(Point(40, 30), o2), True)
        check(pointInShape(Point(0, 0), o2), False)
        check(pointInShape(Point(30, 65), o2), True)
        check(pointInShape(Point(40, 17), o2), True)



import inspect
import sys
from typing import Callable, Dict, Optional, Any

from untypy.error import UntypyAttributeError, UntypyNameError, UntypyTypeError
from untypy.impl.any import SelfChecker
from untypy.impl.choice import ChoiceChecker
from untypy.interfaces import WrappedFunction, TypeChecker, CreationContext, WrappedFunctionContextProvider, \
    ExecutionContext
from untypy.util import ArgumentExecutionContext, ReturnExecutionContext
from untypy.util.typehints import get_type_hints
from untypy.util.condition import FunctionCondition

class FastTypeError(TypeError):
    pass

class TypedFunctionBuilder(WrappedFunction):
    inner: Callable
    signature: inspect.Signature
    _checkers: Optional[Dict[str, TypeChecker]]

    special_args = ['self', 'cls']
    method_name_ignore_return = ['__init__']

    def __init__(self, inner: Callable, ctx: CreationContext):
        self.inner = inner
        self.signature = inspect.signature(inner)
        self.parameters = list(self.signature.parameters.values())
        self.ctx = ctx
        self.fc = None
        self._checkers = None

        try:
            # try to detect errors like missing arguments as early as possible.
            # but not all type annotations are resolvable yet.
            # so ignore `UntypyNameError`s
            self.checkers()
        except UntypyNameError:
            pass

        # The self.fast_sig flags tells wether the signature supports fast matching of arguments.
        # We identified (2022-05-05) that self.wrap_arguments is a performence bottleneck.
        # With type annotations, the performance was factor 8.8 slower than without type
        # annotations
        # So we now have a fastlane for the common case where there are no kw and no variable
        # arguments. For the fastlane, performance is only factor 3.7 slower.
        if hasattr(self.inner, "__fc"):
            self.fc = getattr(self.inner, "__fc")

        self.fast_sig = is_fast_sig(self.parameters, self.fc)

    def checkers(self) -> Dict[str, TypeChecker]:
        if self._checkers is not None:
            return self._checkers

        annotations = get_type_hints(self.inner, self.ctx)

        checkers = {}
        checked_keys = list(self.signature.parameters)

        # Remove self and cls from checking
        if len(checked_keys) > 0 and checked_keys[0] in self.special_args:
            checkers[checked_keys[0]] = SelfChecker()
            checked_keys = checked_keys[1:]

        for key in checked_keys:
            if self.signature.parameters[key].annotation is inspect.Parameter.empty:
                raise self.ctx.wrap(
                    UntypyAttributeError(f"Missing annotation for argument '{key}' of function {self.inner.__name__}\n"
                                         "Partial annotation are not supported."))
            annotation = annotations[key]
            checker = self.ctx.find_checker(annotation)
            if checker is None:
                raise self.ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {annotation}\n"
                                                         f"\tin argument '{key}'"))
            else:
                checkers[key] = checker

        if self.inner.__name__ in self.method_name_ignore_return:
            checkers['return'] = SelfChecker()
        else:
            if not 'return' in annotations:
                annotation = None
            else:
                annotation = annotations['return']
            return_checker = self.ctx.find_checker(annotation)
            if return_checker is None:
                raise self.ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {annotation}\n"
                                                         f"\tin return"))

            checkers['return'] = return_checker

        self._checkers = checkers
        return checkers

    def build(self):
        def wrapper(*args, **kwargs):
            # first is this fn
            caller = sys._getframe(1)
            (args, kwargs, bindings) = self.wrap_arguments(lambda n: ArgumentExecutionContext(self, caller, n), args,
                                                           kwargs)
            ret = self.inner(*args, **kwargs)
            ret = self.wrap_return(args, kwargs, ret, bindings, ReturnExecutionContext(self))
            return ret

        if inspect.iscoroutine(self.inner):
            raise UntypyAttributeError("Async functions are currently not supported.")
        else:
            w = wrapper

        setattr(w, '__wrapped__', self.inner)
        setattr(w, '__name__', self.inner.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)

        # Copy useful attributes
        # This is need for the detection of abstract classes
        for attr in ['__isabstractmethod__']:
            if hasattr(self.inner, attr):
                setattr(w, attr, getattr(self.inner, attr))

        return w

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        return wrap_arguments(self.parameters, self.checkers(), self.signature, self.fc, self.fast_sig,
            ctxprv, args, kwargs)

    def wrap_return(self, args, kwargs, ret, bindings, ctx: ExecutionContext):
        fc_pair = None
        if self.fc is not None:
            fc_pair = (self.fc, bindings)
        return wrap_return(self.checkers()['return'], args, kwargs, ret, fc_pair, ctx)

    def describe(self):
        return str(self.signature)

    def get_original(self):
        return self.inner

    def checker_for(self, name: str) -> TypeChecker:
        return self.checkers()[name]

def _wrap_arguments_fast(parameters, checkers, ctxprv: WrappedFunctionContextProvider, args):
    # Fast case: no kwargs, no rest args, self.fc is None.
    # (self.fc is used for pre- and postconditions, a rarely used feature.)
    # In this case, the order parameter names in args and self.parameters is the
    # same, so we can simply match them by position. As self.fc is None, we do not
    # need to build a inspect.BoundArguments object.
    params = parameters
    n = len(params)
    n_args = len(args)
    if n_args > n:
        raise FastTypeError(f"too many positional arguments")
    wrapped_args = [None] * n
    for i in range(n):
        p = params[i]
        name = params[i].name
        if i < n_args:
            a = args[i]
        else:
            a = p.default
            if a == inspect._empty:
                raise FastTypeError(f"missing a required argument: {name!r}")
        check = checkers[name]
        ctx = ctxprv(name)
        wrapped = check.check_and_wrap(a, ctx)
        wrapped_args[i] = wrapped
    return (wrapped_args, {}, None)

def wrap_arguments(parameters, checkers, signature, fc, fast_sig,
    ctxprv: WrappedFunctionContextProvider, args, kwargs, expectSelf=False):
    if not kwargs and fast_sig:
        try:
            return _wrap_arguments_fast(parameters, checkers, ctxprv, args)
        except FastTypeError as e:
            err = UntypyTypeError(header=str(e))
            if expectSelf and "self" not in parameters:
                err = err.with_note("Hint: 'self'-parameter was omitted in declaration.")
            raise ctxprv("").wrap(err)
    try:
        bindings = signature.bind(*args, **kwargs)
    except TypeError as e:
        err = UntypyTypeError(header=str(e))
        if expectSelf and "self" not in parameters:
            err = err.with_note("Hint: 'self'-parameter was omitted in declaration.")
        raise ctxprv("").wrap(err)
    bindings.apply_defaults()
    if fc is not None:
        fc.prehook(bindings, ctxprv)
    for name in bindings.arguments:
        check = checkers[name]
        ctx = ctxprv(name)
        bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
    return bindings.args, bindings.kwargs, bindings

def is_fast_sig(parameters, fc):
    if fc:
        return False
    fast_sig = True
    for p in parameters:
        # See https://docs.python.org/3/glossary.html#term-parameter for the
        # different kinds of parameters
        if p.kind != inspect._POSITIONAL_ONLY and p.kind != inspect._POSITIONAL_OR_KEYWORD:
            fast_sig = False
            break
    return fast_sig

def wrap_return(check: TypeChecker, args: list[Any], kwds: dict[str, Any], ret: Any,
    fc_pair: tuple[FunctionCondition, inspect.BoundArguments], ctx: ExecutionContext):
    if isinstance(check, ChoiceChecker):
        check = check.get_checker(args, kwds)
    result = check.check_and_wrap(ret, ctx)
    if fc_pair is not None:
        fc, bindings = fc_pair
        fc.posthook(result, bindings, ctx, check)
    return result




from typing import Any, Optional
from untypy.interfaces import TypeCheckerFactory, CreationContext, TypeChecker, ExecutionContext
from untypy.util.debug import debug, isDebug
import untypy.error

# A type depending on the arguments passed to a function. Can only be used as the return type
# of a function.
# Written Choice[T1, ..., Tn, fun]. fun is called with the arguments of the function
# call (possible including self and empty kwargs) and the type checkers for T1, ..., Tn.
# It then returns a type checker.
class Choice:
    def __init__(self, types, select_fun):
        self.types = types
        self.select_fun = select_fun

    def __class_getitem__(cls, args):
        if len(args) < 2:
            raise untypy.error.WyppTypeError(f'Choice requires at least two arguments: Choice[TYPE_1, ..., FUN]')
        return Choice(args[:-1], args[-1])

class ChoiceFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if isinstance(annotation, Choice):
            checkers = []
            for t in annotation.types:
                checker = ctx.find_checker(t)
                if checker is None:
                    return None
                checkers.append(checker)
            return ChoiceChecker(checkers, annotation.select_fun)
        else:
            return None

class ChoiceChecker(TypeChecker):

    def __init__(self, checkers: list[TypeChecker], select_fun):
        self.checkers = checkers
        self.select_fun = select_fun

    def describe(self) -> str:
        l = [c.describe() for c in self.checkers]
        return f"Choice[{', '.join(l)}, <fun>]"

    def may_be_wrapped(self) -> bool:
        return True

    def __illegal_use(self):
        raise untypy.error.WyppTypeError(f"{self.describe()} can only be used as the return type of a function or method")

    def base_type(self) -> list[Any]:
        self.__illegal_use()

    def base_type_priority(self) -> int:
        return 0

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        self.__illegal_use()

    def get_checker(self, args: list[Any], kwds: dict[str, Any]) -> TypeChecker:
        c = self.select_fun(*args, kwds, *self.checkers)
        if isDebug:
            all = [c.describe() for c in self.checkers]
            debug(f'Choice returned checker {c.describe()} for args={args}, kwds={kwds}. All checkers: {",".join(all)}')
        return c



from typing import TypeVar, Generic, Iterable, Optional, Union, Any, Iterator
import typing
from untypy.impl.interfaces.util import overwrite
from untypy.impl.choice import Choice
from untypy.interfaces import CreationContext
from untypy.util import ReturnExecutionContext
from untypy.util.condition import postcondition

I = TypeVar("I")

class List(Generic[I], list):
    # doc @ https://docs.python.org/3/tutorial/datastructures.html
    # and https://docs.python.org/3/library/stdtypes.html#common-sequence-operations
    # Exact signatures are undocumented :/
    # HINT: Argument names must match.

    def append(self, object: I) -> None: pass

    def extend(self, iterable: Iterable[I]) -> None: pass

    def insert(self, i: int, x: I) -> None: pass

    def remove(self, x: I) -> None: pass

    def pop(self, i: int = -1) -> Optional[I]: pass

    def clear(self) -> None: pass

    def index(self, value: I, start: Optional[int] = 0, stop: Optional[int] = 9223372036854775807) -> int:
        # get index of list
        pass

    def count(self, value: I) -> int: pass

    # inner list will check type of key.
    def sort(self, *, key: Any = None, reverse: bool = False) -> None: pass

    def __contains__(self, key: Any) -> bool: pass

    def __delitem__(self, i: Union[int, slice]): pass

    def __getitem__(self, i: Union[int, slice]) -> \
        Choice[I, Any, lambda self, i, kws, ti, tl: ti if isinstance(i, int) else tl]: pass

    def __add__(self, other: Iterable) -> Any: pass

    def __mul__(self, n: int) -> Any: pass

    def __iadd__(self, other: Iterable[I]) -> Any: pass

    def __imul__(self, n: int) -> Any: pass

    def __setitem__(self, key: Union[int, slice], value: Any) -> None: pass

    def __iter__(self) -> Iterator[I]: pass
    # FIXME: complete methods



import typing
import abc
import collections
from untypy.error import UntypyError
from untypy.util.debug import debug

def _f():
    yield 0
generatorType = type(_f())

class WyppWrapError(Exception):
    pass

def patch(self, ty, extra):
    self.__extra__ = extra
    w = self.__wrapped__
    m = None
    if hasattr(w, '__module__'):
        m = getattr(w, '__module__')
    ty.__module__ = m
    try:
        self.__class__ = ty
    except TypeError as e:
        raise WyppWrapError(f'Cannot wrap {self.__wrapped__} of type {type(self.__wrapped__)} ' \
            f'at type {ty}. Original error: {e}')

class WrapperBase:
    def __eq__(self, other):
        if hasattr(other, '__wrapped__'):
            return self.__wrapped__ == other.__wrapped__
        else:
            return self.__wrapped__ == other
    def __ne__(self, other):
        return not self.__eq__(other)
    def __hash__(self):
        return hash(self.__wrapped__)
    def __patch__(self, ms, name=None, extra={}):
        cls = self.__class__
        if name is None:
            name = cls.__name__
        ty = type(name, (cls,), ms)
        patch(self, ty, extra)
    def __repr__(self):
        #w = self.__wrapped__
        #return f"Wrapper(addr=0x{id(self):09x}, wrapped_addr=0x{id(w):09x}, wrapped={repr(w)}"
        return repr(self.__wrapped__)
    def __str__(self):
        return str(self.__wrapped__)
    def __cast__(self, other):
        if type(other) == type(self) and hasattr(other, '__wrapped__'):
            return other.__wrapped__
        else:
            return other
    def __reduce__(self): return self.__wrapped__.__reduce__()
    def __reduce_ex__(self): return self.__wrapped__.__reduce_ex__()
    def __sizeof__(self): return self.__wrapped__.__sizeof__()

class ObjectWrapper(WrapperBase):
    def __init__(self, baseObject):
        self.__dict__ = baseObject.__dict__
        self.__wrapped__ = baseObject
    def __patch__(self, ms, name=None, extra={}):
        cls = self.__class__
        if name is None:
            name = cls.__name__
        wrappedCls = type(self.__wrapped__)
        ty = type(name, (wrappedCls, cls), ms)
        patch(self, ty, extra)

class ABCObjectWrapper(abc.ABC, ObjectWrapper):
    pass

# Superclasses in reverse order.
class ABCObjectWrapperRev(ObjectWrapper, abc.ABC):
    pass

# A wrapper for list such that the class is a subclass of the builtin list class.
# A wrapper for list such that the class is a subclass of the builtin list class.
class ListWrapper(WrapperBase, list): # important: inherit from WrapperBase first
    def __new__(cls, content):
        # the constructor of list copies the list passed to it. Thus, we use an empty list.
        # IMPORTANT: we need to override *all* methods provided by list so that
        # all methods operate on the list being wrapped.
        self = super().__new__(cls, [])
        self.__wrapped__ = content
        return self
    # defined in WrapperBase: __repr__, __str__, __eq__, __hash__
    def __lt__(self, other): return self.__wrapped__.__lt__(self.__cast__(other))
    def __le__(self, other): return self.__wrapped__.__le__(self.__cast__(other))
    def __gt__(self, other): return self.__wrapped__.__gt__(self.__cast__(other))
    def __ge__(self, other): return self.__wrapped__.__ge__(self.__cast__(other))
    def __contains__(self, item): return self.__wrapped__.contains(item)
    def __len__(self): return self.__wrapped__.__len__()
    def __getitem__(self, i): return self.__wrapped__.__getitem__(i)
    def __setitem__(self, i, x): return self.__wrapped__.__setitem__(i, x)
    def __delitem__(self, i): return self.__wrapped__.__delitem__(i)
    def __add__(self, other): return self.__wrapped__.__add__(self.__cast__(other))
    def __radd__(self, other): return self.__cast__(other) + self.__wrapped__
    def __iadd__(self, other): return self.__wrapped__.__iadd__(self.__cast__(other))
    def __mul__(self, n): return self.__wrapped__.__mul__(n)
    __rmul__ = __mul__
    def __imul__(self, n): return self.__wrapped__.__imul__(n)
    def __iter__(self): return self.__wrapped__.__iter__()
    def __copy__(self): return self.__wrapped__.copy()
    def __reversed__(self): return self.__wrapped__.__reversed__()
    def append(self, item): return self.__wrapped__.append(item)
    def extend(self, iter): return self.__wrapped__.extend(iter)
    def insert(self, i, item): return self.__wrapped__.insert(i, item)
    def pop(self, i=-1): return self.__wrapped__.pop(i)
    def remove(self, item): return self.__wrapped__.remove(item)
    def clear(self): return self.__wrapped__.clear()
    def copy(self): return self.__wrapped__.copy()
    def count(self, item): return self.__wrapped__.count(item)
    def index(self, item, *args): return self.__wrapped__.index(item, *args)
    def reverse(self): return self.__wrapped__.reverse()
    def sort(self, /, *args, **kwds): return self.__wrapped__.sort(*args, **kwds)

class SetWrapper(WrapperBase, set):
    def __new__(cls, content):
        self = super().__new__(cls, set())
        self.__wrapped__ = content
        return self
    def add(self, x): return self.__wrapped__.add(x)
    def clear(self): return self.__wrapped__.clear()
    def copy(self): return self.__wrapped__.copy()
    def difference(self, *others): return self.__wrapped__.difference(*others)
    def difference_update(self, *others): return self.__wrapped__.difference_update(*others)
    def discard(self, elem): return self.__wrapped__.discard(elem)
    def intersection(self, *others): return self.__wrapped__.intersection(*others)
    def intersection_update(self, *others): return self.__wrapped__.intersection_update(*others)
    def isdisjoint(self, other): return self.__wrapped__.isdisjoint(other)
    def issubset(self, other): return self.__wrapped__.issubset(other)
    def issuperset(self, other): return self.__wrapped__.issuperset(other)
    def pop(self): return self.__wrapped__.pop()
    def remove(self, elem): return self.__wrapped__.remove(elem)
    def symmetric_difference(self, *others): return self.__wrapped__.symmetric_difference(*others)
    def symmetric_difference_update(self, *others): return self.__wrapped__.symmetric_difference_update(*others)
    def union(self, *others): return self.__wrapped__.union(*others)
    def update(self, *others): return self.__wrapped__.update(*others)
    def __contains__(self, x): return self.__wrapped__.__contains__(x)
    def __le__(self, other): return self.__wrapped__.__le__(self.__cast__(other))
    def __lt__(self, other): return self.__wrapped__.__lt__(self.__cast__(other))
    def __ge__(self, other): return self.__wrapped__.__ge__(self.__cast__(other))
    def __gt__(self, other): return self.__wrapped__.__gt__(self.__cast__(other))
    def __and__(self, other): return self.__wrapped__.__and__(self.__cast__(other))
    def __rand__(self, other): return self.__wrapped__.__rand__(self.__cast__(other))
    def __iand__(self, other): return self.__wrapped__.__iand__(self.__cast__(other))
    def __ior__(self, other): return self.__wrapped__.__ior__(self.__cast__(other))
    def __ixor__(self, other): return self.__wrapped__.__ixor__(self.__cast__(other))
    def __or__(self, other): return self.__wrapped__.__or__(self.__cast__(other))
    def __ror__(self, other): return self.__wrapped__.__ror__(self.__cast__(other))
    def __rxor__(self, other): return self.__wrapped__.__rxor__(self.__cast__(other))
    def __xor__(self, other): return self.__wrapped__.__xor__(self.__cast__(other))
    def __rsub__(self, other): return self.__wrapped__.__rsub__(self.__cast__(other))
    def __sub__(self, other): return self.__wrapped__.__sub__(self.__cast__(other))
    def __isub__(self, other): return self.__wrapped__.__isub__(self.__cast__(other))
    def __iter__(self): return self.__wrapped__.__iter__()
    def __len__(self): return self.__wrapped__.__len__()

class DictWrapper(WrapperBase, dict):
    def __new__(cls, content):
        self = super().__new__(cls, {})
        self.__wrapped__ = content
        return self
    def __len__(self): return self.__wrapped__.__len__()
    def __getitem__(self, key): return self.__wrapped__.__getitem__(key)
    def __setitem__(self, key, item): return self.__wrapped__.__setitem__(key, item)
    def __delitem__(self, key): return self.__wrapped__.__delitem__(key)
    def __iter__(self): return self.__wrapped__.__iter__()
    def __contains__(self, key): return self.__wrapped__.__contains__(key)
    def __or__(self, other): return self.__wrapped__.__or__(self.__cast__(other))
    def __ror__(self, other): return self.__wrapped__.__ror__(self.__cast__(other))
    def __ior__(self, other): return self.__wrapped__.__ior__(self.__cast__(other))
    def __copy__(self): return self.__wrapped__.__copy__()
    def __lt__(self, other): return self.__wrapped__.__lt__(self.__cast__(other))
    def __le__(self, other): return self.__wrapped__.__le__(self.__cast__(other))
    def __gt__(self, other): return self.__wrapped__.__gt__(self.__cast__(other))
    def __ge__(self, other): return self.__wrapped__.__ge__(self.__cast__(other))
    def copy(self): return self.__wrapped__.copy()
    def __reversed__(self): return self.__wrapped__.__reversed__()
    __marker = object()
    def pop(self, key, default=__marker):
        if default == self.__marker:
            return self.__wrapped__.pop(key)
        else:
            return self.__wrapped__.pop(key, default)
    def popitem(self): return self.__wrapped__.popitem()
    def clear(self): return self.__wrapped__.clear()
    def update(self, other=(), /, **kwds): return self.__wrapped__.update(self.__cast__(other), **kwds)
    def setdefault(self, key, default=None): return self.__wrapped__.setdefault(key, default=default)
    def get(self, key, default=None): return self.__wrapped__.get(key, default)
    def keys(self): return self.__wrapped__.keys()
    def items(self): return self.__wrapped__.items()
    def values(self): return self.__wrapped__.values()

# Tuple and string wrapper are simpler because these types are immutable
class StringWrapper(str, WrapperBase):
    def __new__(cls, content):
        self = super().__new__(cls, content)
        self.__wrapped__ = content
        return self

class TupleWrapper(tuple, WrapperBase):
    def __new__(cls, content):
        self = super().__new__(cls, content)
        self.__wrapped__ = content
        return self

# These methods are not delegated to the wrapped object
_blacklist = [
    '__class__', '__delattr__', '__dict__', '__dir__', '__doc__',
    '__getattribute__', '__get_attr_', '__init_subclass__'
    '__init__', '__new__', '__del__', '__repr__', '__setattr__', '__str__',
    '__hash__', '__eq__', '__patch__',
    '__class_getitem__',  '__subclasshook__']

_extra = ['__next__']

# SimpleWrapper is a fallback for types that cannot be used as base types
class SimpleWrapper(WrapperBase):
    def __init__(self, baseObject):
        self.__wrapped__ = baseObject
    def __patch__(self, ms, name=None, extra={}):
        cls = self.__class__
        if name is None:
            name = cls.__name__
        baseObject = self.__wrapped__
        for x in dir(baseObject) + _extra:
            if x not in ms and x not in _blacklist and hasattr(baseObject, x):
                ms[x] = getattr(baseObject, x)
        ty = type(name, (cls,), ms) #
        patch(self, ty, extra)

class ValuesViewWrapper(SimpleWrapper):
    pass
collections.abc.ValuesView.register(ValuesViewWrapper)

class ItemsViewWrapper(SimpleWrapper):
    pass
collections.abc.ItemsView.register(ItemsViewWrapper)

class KeysViewWrapper(SimpleWrapper):
    pass
collections.abc.KeysView.register(KeysViewWrapper)

def wrap(obj, methods, name=None, extra={}, simple=False):
    if simple:
        w = SimpleWrapper(obj)
    elif isinstance(obj, list):
        w = ListWrapper(obj)
    elif isinstance(obj, tuple):
        w = TupleWrapper(obj)
    elif isinstance(obj, dict):
        w = DictWrapper(obj)
    elif isinstance(obj, str):
        w = StringWrapper(obj)
    elif isinstance(obj, set):
        w = SetWrapper(obj)
    elif isinstance(obj, collections.abc.ValuesView):
        w = ValuesViewWrapper(obj)
    elif isinstance(obj, collections.abc.KeysView):
        w = KeysViewWrapper(obj)
    elif isinstance(obj, collections.abc.ItemsView):
        w = ItemsViewWrapper(obj)
    elif isinstance(obj, typing.Generic):
        w = SimpleWrapper(obj)
    elif isinstance(obj, generatorType):
        w = SimpleWrapper(obj)
    elif isinstance(obj, abc.ABC) and hasattr(obj, '__dict__'):
        try:
            w = ABCObjectWrapper(obj)
        except WyppWrapError:
            try:
                w = ABCObjectWrapperRev(obj)
            except WyppWrapError:
                w = SimpleWrapper(obj)
    elif hasattr(obj, '__dict__'):
        w = ObjectWrapper(obj)
    else:
        w = SimpleWrapper(obj)
    w.__patch__(methods, name, extra)
    wname = name
    if wname is None:
        wname = str(type(w))
    debug(f"Wrapping {obj} at 0x{id(obj):09x} as {wname}, simple={simple}, wrapper=0x{id(w):09x}")
    return w



import typing
from collections.abc import Iterable as ABCIterable
from collections.abc import Iterator as ABCIterator
from collections.abc import Sequence as ABCSequence
from typing import Optional, Any

from untypy.error import UntypyAttributeError
from untypy.impl.interfaces.iterable import Iterable, Iterator
from untypy.impl.interfaces.sequence import Sequence
from untypy.impl.interfaces.dict import Dict, DictLike
from untypy.impl.interfaces.list import List
from untypy.impl.interfaces.set import Set
from untypy.impl.protocol import ProtocolChecker
from untypy.interfaces import TypeCheckerFactory, TypeChecker, CreationContext

InterfaceMapping = {
    dict: Dict,
    typing.Dict: Dict,
    DictLike: DictLike,
    list: List,
    typing.List: List,
    set: Set,
    typing.Set: Set,
    ABCIterable: Iterable,
    typing.Iterable: Iterable,
    typing.Iterator: Iterator,
    ABCIterator: Iterator,
    ABCSequence: Sequence,
    typing.Sequence: Sequence
}

class InterfaceFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext, omit_tyargs=False) -> Optional[TypeChecker]:
        if annotation in InterfaceMapping:
            # Assume Any if no parameters are given
            protocol = InterfaceMapping[annotation]
            bindings = protocol.__parameters__
            if len(bindings) == 0:
                raise AssertionError(f"This is a BUG. {annotation} has no generic params.")
            anys = (Any, ) * len(bindings)
            # handle Python inconsistency
            if hasattr(annotation, '__class_getitem__'):
                return self.create_from(
                    annotation.__class_getitem__(anys),
                    ctx,
                    omit_tyargs=True
                )
            elif hasattr(annotation, '__getitem__'):
                return self.create_from(
                    annotation.__getitem__(anys),
                    ctx,
                    omit_tyargs=True
                )

        elif hasattr(annotation, '__origin__') and \
            hasattr(annotation, '__args__') and annotation.__origin__ in InterfaceMapping:
            protocol = InterfaceMapping[annotation.__origin__]
            bindings = protocol.__parameters__  # args of Generic super class
            origin = annotation.__origin__
            inner_checkers = []
            for param in annotation.__args__:
                ch = ctx.find_checker(param)
                if ch is None:
                    raise UntypyAttributeError(f"Could not resolve annotation {param} inside of {annotation}")
                inner_checkers.append(ch)
            if len(inner_checkers) != len(bindings):
                raise UntypyAttributeError(f"Expected {len(bindings)} type arguments inside of {annotation}")
            if omit_tyargs:
                name = f"{origin.__name__}"
            else:
                name = f"{origin.__name__}[" + (', '.join(map(lambda t: t.describe(), inner_checkers))) + "]"
            bindings = dict(zip(bindings, inner_checkers))
            ctx = ctx.with_typevars(bindings)
            return ProtocolChecker(protocol, ctx, altname=name, omit_tyargs=omit_tyargs, ty=origin)

        # Non Generic
        elif annotation in InterfaceMapping:
            protocol = InterfaceMapping[annotation]
            # Edge-Case `TypingSequence has no __name__, like every other class`
            if annotation == typing.Sequence:
                name = 'Sequence'
            else:
                name = annotation.__name__
            return ProtocolChecker(protocol, ctx, altname=name, omit_tyargs=omit_tyargs, ty=annotation)
        else:
            return None



import unittest
from writeYourProgram import *
from drawingLib import *

setDieOnCheckFailures(True)

class TestDrawing(unittest.TestCase):

    def test_drawing(self):
        shapes = [
            FixedShape('rectangle', 'red', Size(3, 1), Point(-1, 1.5)),
            FixedShape('ellipsis', 'blue', Size(2, 3), Point(-1, -0.5)),
            FixedShape('rectangle', 'black', Size(2, 2), Point(1.5, 0))
        ]
        # Do not draw, does not work with CI
        # drawFixedShapes(shapes, stopAfter=0.5)



import unittest
import sample
from writeYourProgram import *
from writeYourProgram import deepEq
import writeYourProgram as wypp

wypp.setDieOnCheckFailures(True)

class C:
    def __init__(self, x):
        self.x = x

class D:
    def __init__(self, x):
        self.x = x

class A:
    def __init__(self, x):
        self.x = x
    def __eq__(self, other):
        return self is other

@record
class Point:
    x: float
    y: float

class TestDeepEq(unittest.TestCase):
    def test_eq(self):
        self.assertTrue(deepEq(1, 1.0, floatEqWithDelta=True))
        self.assertTrue(deepEq(1, 1.0))
        self.assertFalse(deepEq(1, 1.01, floatEqWithDelta=True))
        self.assertTrue(deepEq(1, 1.00000001, floatEqWithDelta=True))
        self.assertTrue(deepEq(1.0, 1.00000001, floatEqWithDelta=True))
        self.assertFalse(deepEq([1,2,3], [1,2]))
        self.assertTrue(deepEq('foo bar', 'foo ' + 'bar'))
        self.assertFalse(deepEq("foo", "bar"))
        self.assertFalse(deepEq('x', 'y'))
        self.assertTrue(deepEq([], []))
        self.assertFalse(deepEq([1], [2]))
        self.assertFalse(deepEq([1, 'foo'], [1.000000000000001, 'foo']))
        self.assertTrue(deepEq([1, 'foo'], [1.000000000000001, 'foo'], floatEqWithDelta=True))
        self.assertTrue(deepEq([(42.0, 32.0000000001)], [(42.00000000000000001, 32)],
                               floatEqWithDelta=True))
        self.assertTrue(deepEq({"foo": 42.0}, {"foo": 42.000000000000000001},
                               floatEqWithDelta=True))
        self.assertFalse(deepEq({"foo": 42.0}, {"foo": 42.1}, floatEqWithDelta=True))
        self.assertTrue(deepEq({"foo": [42.0]}, {"foo": [42.000000000000000001]},
                        floatEqWithDelta=True))
        self.assertTrue(deepEq({"foo": {'bar': [42.0]}}, {"foo": {'bar': [42.000000000000000001]}},
                               floatEqWithDelta=True))
        l1 = [1]
        l2 = [1]
        l1.append(l1)
        l2.append(l2)
        try:
            deepEq(l1, l2)
            self.fail("Exception expected")
        except RecursionError:
            pass
        self.assertFalse(deepEq(C(42.0), C(42.0000000000001),
                                structuralObjEq=False, floatEqWithDelta=True))
        self.assertTrue(deepEq(C(42.0), C(42.0000000000001),
                               structuralObjEq=True, floatEqWithDelta=True))
        self.assertTrue(deepEq(C(42.0), D(42.0000000000001),
                               structuralObjEq=True, floatEqWithDelta=True))
        self.assertFalse(deepEq(C(42.0), D(42.0000000000001),
                                structuralObjEq=False, floatEqWithDelta=True))
        self.assertTrue(deepEq(
            sample.Point(1.0, 2.0), sample.Point(1.00000000001, 2),
            structuralObjEq=True, floatEqWithDelta=True
        ))
        self.assertFalse(deepEq(
            sample.Point(1.0, 2.0), sample.Point(1.00000000001, 2),
            structuralObjEq=False, floatEqWithDelta=True
        ))
        self.assertTrue(deepEq(
            Point(1.0, 2.0), sample.Point(1.0, 2.0),
            structuralObjEq=True, floatEqWithDelta=True)
        )
        self.assertFalse(deepEq(
            Point(1.0, 2.0), sample.Point(1.0, 2.0),
            structuralObjEq=False, floatEqWithDelta=True)
        )
        self.assertFalse(deepEq(
            sample.Point(1.0, 3.0), sample.Point(1.00000000001, 2),
            structuralObjEq=True, floatEqWithDelta=True
        ))
        self.assertFalse(deepEq(
            sample.Point(1.0, 3.0), sample.Point(1.00000000001, 2),
            structuralObjEq=False, floatEqWithDelta=True
        ))
        self.assertTrue(deepEq(A(2), A(2), structuralObjEq=True, floatEqWithDelta=True))
        self.assertFalse(deepEq(A(2), A(2), structuralObjEq=False, floatEqWithDelta=True))
        self.assertTrue(deepEq(foo, foo))
        self.assertTrue(deepEq(foo, foo, structuralObjEq=True))
        self.assertFalse(deepEq(foo, bar))
        self.assertFalse(deepEq(foo, bar, structuralObjEq=True))

def foo():
    pass

def bar():
    return 1



from typing import Generic, TypeVar, Optional, Tuple, Iterable, Any, Iterator
from untypy.impl.interfaces.iterable import OnlyIterable

I = TypeVar("I")

class Set(Generic[I]):
    def add(self, other: I) -> None: pass
    def clear(self) -> None: pass
    def copy(self) -> Any: pass
    def difference(self, *others: Tuple[Iterable, ...]) -> Any: pass
    def difference_update(self, *others: Tuple[Iterable[I], ...]) -> None: pass
    def discard(self, elem: I) -> None: pass
    def intersection(self, *others: Tuple[Iterable, ...]) -> Any: pass
    def intersection_update(self, *others: Tuple[Iterable[I], ...]) -> None: pass
    def isdisjoint(self, other: set) -> bool: pass
    def issubset(self, other: set) -> bool: pass
    def issuperset(self, other: set) -> bool: pass
    def pop(self) -> Optional[I]: pass
    def remove(self, elem: I) -> None: pass
    def symmetric_difference(self, *others: Tuple[Iterable, ...]) -> Any: pass
    def symmetric_difference_update(self, *others: Tuple[Iterable[I], ...]) -> None: pass
    def union(self, *others: Tuple[Iterable, ...]) -> Any: pass

    # Using OnlyIterable here makes no other methods than the one provide in Iterable
    # available. This is required because the implementation of update has shortcuts
    # bypassing type checks if an element is of type set.
    def update(self, *others: Tuple[OnlyIterable[I], ...]) -> None:
        pass

    def __contains__(self, x: Any) -> bool: pass
    def __delattr__(self, name: str) -> None: pass

    def __le__(self, other: Any) -> bool: pass
    def __lt__(self, other: Any) -> bool: pass
    def __ge__(self, other: Any) -> bool: pass
    def __gt__(self, other: Any) -> bool: pass

    def __and__(self, other: set) -> Any: pass
    def __rand__(self, other: set) -> Any: pass
    def __iand__(self, other: set) -> Any: pass
    def __ior__(self, other: set) -> Any: pass # FIXME: result should be set[I]
    def __isub__(self, other: set) -> Any: pass
    def __ixor__(self, other: set) -> Any: pass
    def __or__(self, other: set) -> Any: pass
    def __ror__(self, other: set) -> Any: pass
    def __rxor__(self, other: set) -> Any: pass
    def __xor__(self, other: set) -> Any: pass
    def __rsub__(self, other: set) -> Any: pass
    def __sub__(self, other: set) -> Any: pass

    def __iter__(self) -> Iterator[I]: pass
    def __len__(self) -> int: pass
    def __setattr__(self, name: str, x: Any) -> None: pass



import inspect
from collections import namedtuple
from types import FunctionType
from typing import Callable

from untypy.error import Location
from untypy.impl import DefaultCreationContext
from untypy.interfaces import WrappedFunction
from untypy.util.typedfunction import TypedFunctionBuilder

Config = namedtuple('PatchConfig', ['verbose', 'checkedprefixes'])
DefaultConfig = Config(verbose=False, checkedprefixes=[""])
not_patching = ['__class__']

GlobalPatchedList = set()


def patch_class(clas: type, cfg: Config):
    if clas in GlobalPatchedList:
        return clas
    GlobalPatchedList.add(clas)

    try:
        ctx = DefaultCreationContext(
            typevars=dict(),
            declared_location=Location(
                file=inspect.getfile(clas),
                line_no=inspect.getsourcelines(clas)[1],
                line_span=len(inspect.getsourcelines(clas)[0]),
            ), checkedpkgprefixes=cfg.checkedprefixes)
    except (TypeError, OSError) as e:  # Built in types
        ctx = DefaultCreationContext(
            typevars=dict(),
            declared_location=Location(
                file="<not found>",
                line_no=0,
                line_span=1
            ), checkedpkgprefixes=cfg.checkedprefixes,
        )

    setattr(clas, '__patched', True)

def wrap_function(fn: FunctionType, cfg: Config) -> Callable:
    if len(inspect.getfullargspec(fn).annotations) > 0:
        if cfg.verbose:
            print(f"Patching Function: {fn.__name__}")
        return TypedFunctionBuilder(fn, DefaultCreationContext(
            typevars=dict(),
            declared_location=WrappedFunction.find_location(fn),
            checkedpkgprefixes=cfg.checkedprefixes, eval_context=fn.__globals__)).build()
    else:
        return fn


def wrap_class(a: type, cfg: Config) -> Callable:
    return WrappedType(a, DefaultCreationContext(
        typevars=dict(),
        declared_location=Location.from_code(a),
        checkedpkgprefixes=cfg.checkedprefixes))



import abc
import inspect
import sys
import typing
from typing import Protocol, Any, Optional, Callable, Union, TypeVar, Dict, Tuple

import untypy.util.display as display
from untypy.error import UntypyTypeError, UntypyAttributeError, Frame, Location, ResponsibilityType
from untypy.impl.any import SelfChecker, AnyChecker
from untypy.interfaces import TypeCheckerFactory, CreationContext, TypeChecker, ExecutionContext, \
    WrappedFunctionContextProvider
from untypy.util import WrappedFunction, ArgumentExecutionContext, ReturnExecutionContext
from untypy.util.condition import FunctionCondition
from untypy.util.typehints import get_type_hints
import untypy.util.wrapper as wrapper
import untypy.util.typedfunction as typedfun

class ProtocolFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if isinstance(annotation, type) and Protocol in annotation.mro():
            return ProtocolChecker(annotation, ctx)
        elif hasattr(annotation, '__args__') and \
            hasattr(annotation, '__origin__') and \
            hasattr(annotation.__origin__, '__mro__') and \
            typing.Protocol in annotation.__origin__.__mro__:
            return ProtocolChecker(annotation, ctx)
        else:
            return None


def _find_bound_typevars(clas: type) -> Tuple[type, Dict[TypeVar, Any]]:
    if not hasattr(clas, '__args__') or not hasattr(clas, '__origin__'):
        return (clas, dict())
    if not hasattr(clas.__origin__, '__parameters__'):
        return (clas, dict())

    keys = clas.__origin__.__parameters__
    values = clas.__args__

    if len(keys) != len(values):
        raise UntypyAttributeError(f"Some unbound Parameters in {clas.__name__}. "
                                   f"keys={keys} do not match values={values}.",
                                   [Location(
                                       file=inspect.getfile(clas),
                                       line_no=inspect.getsourcelines(clas)[1],
                                       line_span=len(inspect.getsourcelines(clas)[0]))])
    return (clas.__origin__, dict(zip(keys, values)))


def get_proto_members(proto: type, ctx: CreationContext) -> dict[
    str, Tuple[inspect.Signature, dict[str, TypeChecker], FunctionCondition]]:
    blacklist = ['__init__', '__class__', '__delattr__', '__dict__', '__dir__',
                 '__doc__', '__getattribute__', '__getattr__', '__init_subclass__',
                 '__new__', '__setattr__', '__subclasshook__', '__weakref__',
                 '__abstractmethods__', '__class_getitem__']

    member_dict = {}
    for [name, member] in inspect.getmembers(proto):
        if name in blacklist:
            continue

        if inspect.isfunction(member):
            member = WrappedFunction.find_original(member)
            signature = inspect.signature(member)

            is_typed = len(inspect.getfullargspec(member).annotations) != 0

            checkers = {}
            if not is_typed:
                # Use Any for any type
                for key in signature.parameters:
                    if key == 'self':
                        checkers[key] = SelfChecker()
                    else:
                        checkers[key] = AnyChecker()
                checkers['return'] = AnyChecker()
            else:
                annotations = get_type_hints(member, ctx)
                for key in signature.parameters:
                    if key == 'self':
                        checkers[key] = SelfChecker()
                    else:
                        param = signature.parameters[key]
                        if param.annotation is inspect.Parameter.empty:
                            raise ctx.wrap(UntypyAttributeError(
                                f"Missing annotation for argument '{key}' of function {member.__name__} "
                                f"in protocol {proto.__name__}\n"))

                        param_anot = annotations[key]
                        if param_anot is proto:
                            checker = SimpleInstanceOfChecker(proto, None)
                        else:
                            checker = ctx.find_checker(param_anot)
                        if checker is None:
                            raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {param.annotation}\n"
                                                                f"for argument '{key}' of function {member.__name__} "
                                                                f"in protocol {proto.__name__}.\n"))
                        checkers[key] = checker

                if signature.return_annotation is inspect.Parameter.empty:
                    return_annotation = None
                else:
                    return_annotation = annotations['return']
                if return_annotation is proto:  # Self as Return Type would led to endless recursion
                    return_checker = SimpleInstanceOfChecker(proto, None)
                else:
                    return_checker = ctx.find_checker(return_annotation)

                if return_checker is None:
                    raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {signature.return_annotation}\n"
                                                        f"for return value of function {member.__name__} "
                                                        f"in protocol-like {proto.__name__}.\n"))
                checkers['return'] = return_checker

            fc = None
            if hasattr(member, '__fc'):
                fc = getattr(member, '__fc')
            member_dict[name] = (signature, checkers, fc)
    return member_dict


class ProtocolChecker(TypeChecker):
    def __init__(self, annotation: type, ctx: CreationContext, *, altname : Optional[str] = None,
            omit_tyargs=False, ty: Optional[type]=None):
        (proto, typevars) = _find_bound_typevars(annotation)
        self.ctx = ctx.with_typevars(typevars)
        self.proto = proto
        self._members = None
        self.typevars = typevars
        self.altname = altname
        self.omit_tyargs = omit_tyargs
        self.ty = ty

    @property
    def members(self):
        if not self._members:
            self._members = get_proto_members(self.proto, self.ctx)
        return self._members

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if hasattr(arg, '__wrapped__'):
            # no double wrapping
            arg = getattr(arg, '__wrapped__')
        simple = False
        if hasattr(self.proto, '__protocol_only__'):
            simple = getattr(self.proto, '__protocol_only__')
        if self.ty is not None and not isinstance(arg, self.ty):
            err = ctx.wrap(UntypyTypeError(
                expected=self.describe(),
                given=arg
            ))
            raise err
        return wrapForProtocol(self, arg, self.members, ctx, simple=simple)

    def base_type(self) -> list[Any]:
        # Prevent Classes implementing multiple Protocols in one Union by accident.
        if self.ty:
            return [self.ty]
        else:
            return [Protocol]

    def describe(self) -> str:
        if self.altname is not None:
            return self.altname

        desc = set([])
        if not self.omit_tyargs:
            for name in self.members:
                (sig, binds, cond) = self.members[name]
                for argname in sig.parameters:
                    if isinstance(sig.parameters[argname].annotation, TypeVar):
                        desc.add(binds[argname].describe())
                if isinstance(sig.return_annotation, TypeVar):
                    desc.add(binds['return'].describe())
        if len(desc) > 0:
            # FIXME: what about the ordering of tyvars?
            return f"{self.proto.__name__}[" + (', '.join(desc)) + "]"
        else:
            return f"{self.proto.__name__}"

    def protocol_type(self) -> str:
        return f"protocol"

    def protoname(self):
        return self.describe()

def isInternalProtocol(p: Any):
    if isinstance(p, ProtocolChecker):
        p = p.proto
    if hasattr(p, '__module__'):
        return 'untypy.' in p.__module__
    else:
        return False

def protoMismatchErrorMessage(what: str, proto: Any):
    if isinstance(proto, ProtocolChecker):
        kind = proto.protocol_type()
        name = proto.protoname()
    else:
        kind = 'protocol'
        name = proto.__name__
    isUserDefined = True
    if isInternalProtocol(proto):
        isUserDefined = False
    if isUserDefined:
        return f"{what} does not implement {kind} {name}"
    else:
        return f"{what} is not {display.withIndefiniteArticle(name)}"

def wrapForProtocol(protocolchecker: ProtocolChecker,
                    originalValue: Any,
                    members: dict[str, Tuple[inspect.Signature, dict[str, TypeChecker], FunctionCondition]],
                    ctx: ExecutionContext,
                    simple: bool):
    list_of_attr = dict()
    original = type(originalValue)
    for fnname in members:
        if not hasattr(original, fnname):
            err = ctx.wrap(UntypyTypeError(
                expected=protocolchecker.describe(),
                given=originalValue
            )).with_header(
                protoMismatchErrorMessage(original.__name__, protocolchecker.proto)
            )
            missing = []
            for fnname in members:
                if not hasattr(original, fnname):
                    missing.append(fnname)
            if len(missing) == 2:
                err = err.with_note(f"It is missing the functions '{missing[0]}' and '{missing[1]}'")
            elif len(missing) == 1:
                err = err.with_note(f"It is missing the function '{missing[0]}'")
            raise err

        original_fn = getattr(original, fnname)
        try:
            # fails on built ins - YEAH
            original_fn_signature = inspect.signature(original_fn)
        except:
            original_fn_signature = None

        if hasattr(original_fn, '__wf'):
            original_fn = getattr(original_fn, '__wf')
        (sig, baseArgDict, fc) = members[fnname]

        if original_fn_signature is not None:
            err = None
            if len(sig.parameters) > len(original_fn_signature.parameters):
                err = f"The signature of '{fnname}' does not match. Missing required parameters."
            # Special check for self
            if 'self' in sig.parameters and 'self' not in original_fn_signature.parameters:
                err = f"The signature of '{fnname}' does not match. Missing required parameter self."
            if err is not None:
                raise ctx.wrap(UntypyTypeError(
                    expected=protocolchecker.describe(),
                    given=originalValue
                )).with_header(
                    protoMismatchErrorMessage(original.__name__, protocolchecker.proto)
                ).with_note(err)
            paramDict = dict(zip(original_fn_signature.parameters, sig.parameters))
        else:
            paramDict = {}
            for k in sig.parameters:
                paramDict[k] = k

        list_of_attr[fnname] = ProtocolWrappedFunction(original_fn,
                                                       sig,
                                                       baseArgDict,
                                                       protocolchecker,
                                                       paramDict,
                                                       fc).build()

    name = f"WyppTypeCheck({original.__name__}, {protocolchecker.describe()})"
    return wrapper.wrap(originalValue, list_of_attr, name, extra={'ctx': ctx}, simple=simple)

def _getWrappedName(fn):
    if hasattr(fn, '__name__'):
        return fn.__name__
    elif hasattr(fn, 'fget'):
        return fn.fget.__name__
    else:
        raise ValueError(f'Cannot determine name of {fn}')

def _isProperty(fn):
    return isinstance(fn, property)

class ProtocolWrappedFunction(WrappedFunction):

    def __init__(self,
                 inner: Union[Callable, WrappedFunction],
                 signature: inspect.Signature,
                 checker: Dict[str, TypeChecker], # maps argument names from the protocol to checkers
                 protocol: ProtocolChecker,
                 baseArgs: Dict[str, str], # maps arguments names of the implementing class to argument names of the protocol
                 fc: FunctionCondition):
        self.inner = inner
        self.signature = signature
        self.parameters = list(self.signature.parameters.values())
        self.checker = checker
        self.baseArgs = baseArgs
        self.protocol = protocol
        self.fc = fc
        self.fast_sig = typedfun.is_fast_sig(self.parameters, self.fc)

    def build(self):
        fn = WrappedFunction.find_original(self.inner)
        name = _getWrappedName(fn)
        fn_of_protocol = getattr(self.protocol.proto, name)
        if hasattr(fn_of_protocol, '__wf'):
            fn_of_protocol = getattr(fn_of_protocol, '__wf')

        def wrapper(me, *args, **kwargs):
            inner_object = me.__wrapped__
            inner_ctx = me.__extra__['ctx']

            caller = sys._getframe(1)
            (args, kwargs, bind1) = self.wrap_arguments(lambda n: ArgumentExecutionContext(fn_of_protocol, caller, n),
                                                        (inner_object, *args), kwargs)
            if isinstance(self.inner, WrappedFunction):
                (args, kwargs, bind2) = self.inner.wrap_arguments(
                    lambda n: ProtocolArgumentExecutionContext(self, self.baseArgs[n], n,
                                                               inner_object,
                                                               inner_ctx),
                    args, kwargs)
            if _isProperty(fn):
                ret = fn
            else:
                ret = fn(*args, **kwargs)
            if isinstance(self.inner, WrappedFunction):
                ret = self.inner.wrap_return(args, kwargs, ret, bind2,
                    ProtocolReturnExecutionContext(self, ResponsibilityType.IN, inner_object, inner_ctx))
            return self.wrap_return(args, kwargs, ret, bind1,
                ProtocolReturnExecutionContext(self, ResponsibilityType.OUT, inner_object, inner_ctx))

        async def async_wrapper(*args, **kwargs):
            raise AssertionError("Not correctly implemented see wrapper")

        if inspect.iscoroutine(self.inner):
            w = async_wrapper
        else:
            w = wrapper

        setattr(w, '__wrapped__', fn)
        setattr(w, '__name__', name)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)
        return w

    def get_original(self):
        return self.inner

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        return typedfun.wrap_arguments(self.parameters, self.checker, self.signature,
            self.fc, self.fast_sig, ctxprv, args, kwargs, expectSelf=True)

    def wrap_return(self, args, kwargs, ret, bindings, ctx: ExecutionContext):
        fc_pair = None
        if self.fc is not None:
            fc_pair = (self.fc, bindings)
        return typedfun.wrap_return(self.checker['return'], args, kwargs, ret, fc_pair, ctx)

    def describe(self) -> str:
        fn = WrappedFunction.find_original(self.inner)
        return f"{fn.__name__}" + str(self.signature)

    def checker_for(self, name: str) -> TypeChecker:
        if name == 'return':
            k = 'return'
        else:
            k = self.baseArgs[name]
        return self.checker[k]

    def declared(self) -> Location:
        fn = WrappedFunction.find_original(self.inner)
        return WrappedFunction.find_location(getattr(self.protocol.proto, fn.__name__))


class ProtocolReturnExecutionContext(ExecutionContext):
    def __init__(self, wf: ProtocolWrappedFunction, invert: ResponsibilityType, me: Any, ctx: ExecutionContext):
        self.wf = wf
        self.invert = invert
        self.me = me
        self.ctx = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        err = ReturnExecutionContext(self.wf).wrap(err)

        if err.responsibility_type is self.invert:
            return err
        responsable = WrappedFunction.find_location(self.wf)
        (decl, ind) = err.next_type_and_indicator()
        err = err.with_inverted_responsibility_type()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=self.wf.declared(),
            responsable=responsable
        ))

        inner = self.wf.inner
        if isinstance(inner, WrappedFunction):
            err = err.with_note(
                f"The return value of method '{WrappedFunction.find_original(self.wf).__name__}' does violate the {self.wf.protocol.protocol_type()} '{self.wf.protocol.proto.__name__}'.")
            err = err.with_note(
                f"The annotation '{inner.checker_for('return').describe()}' is incompatible with the {self.wf.protocol.protocol_type()}'s annotation '{self.wf.checker_for('return').describe()}'\nwhen checking against the following value:")

        previous_chain = UntypyTypeError(
            self.me,
            f"{self.wf.protocol.protoname()}"
        ).with_header(
            protoMismatchErrorMessage(type(self.me).__name__, self.wf.protocol)
        )

        previous_chain = self.ctx.wrap(previous_chain)
        if isInternalProtocol(self.wf.protocol):
            return previous_chain
        else:
            return err.with_previous_chain(previous_chain)



class ProtocolArgumentExecutionContext(ExecutionContext):
    def __init__(self, wf: ProtocolWrappedFunction,
                 base_arg: str, this_arg: str, me: Any, ctx: ExecutionContext):
        self.wf = wf
        self.base_arg = base_arg
        self.this_arg = this_arg
        self.me = me
        self.ctx = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        protocol = self.wf.protocol.proto

        (original_expected, _ind) = err.next_type_and_indicator()
        err = ArgumentExecutionContext(self.wf, None, self.base_arg).wrap(err)

        responsable = WrappedFunction.find_location(self.wf)

        (decl, ind) = err.next_type_and_indicator()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=self.wf.declared(),
            responsable=responsable
        ))

        base_expected = self.wf.checker_for(self.this_arg).describe()
        if self.base_arg == self.this_arg:
            err = err.with_note(
                f"Argument {self.this_arg} of method {WrappedFunction.find_original(self.wf).__name__} "
                f"violates the type declared by the "
                f"{self.wf.protocol.protocol_type()} {self.wf.protocol.proto.__name__}.")
        else:
            err = err.with_note(
                f"Argument {self.this_arg} of method {WrappedFunction.find_original(self.wf).__name__} "
                f"violates the type declared for {self.base_arg} in "
                f"{self.wf.protocol.protocol_type()} {self.wf.protocol.proto.__name__}.")
        err = err.with_note(
            f"Annotation {original_expected} is incompatible with the "
            f"{self.wf.protocol.protocol_type()}'s annotation "
            f"{base_expected}.")

        previous_chain = UntypyTypeError(
            self.me,
            f"{self.wf.protocol.protoname()}"
        ).with_header(
            protoMismatchErrorMessage(type(self.me).__name__, self.wf.protocol)
        )

        # Protocols can either be declared explicit or implicit.
        if protocol in type(self.me).__mro__:
            # If it is declared explicit (e.g. Inheritance) the
            # declaration of the inheritance has to be blamed.
            previous_chain = previous_chain.with_frame(Frame(
                # /- Could also be `self.wf.describe()`, which would put "right" signature as "context:".
                # v  But this info may be better suited in the note.
                *previous_chain.next_type_and_indicator(),
                declared=Location.from_code(type(self.me)),
                responsable=Location.from_code(type(self.me))
            )).with_frame(Frame(
                *previous_chain.next_type_and_indicator(),
                declared=self.wf.declared(),
                responsable=responsable
            ))
        else:
            # Else: We need to explain how this protocol was declared.
            previous_chain = self.ctx.wrap(previous_chain)

        if isInternalProtocol(self.wf.protocol):
            return previous_chain
        else:
            return err.with_previous_chain(previous_chain)


class SimpleInstanceOfChecker(TypeChecker):
    def __init__(self, annotation: type, ctx: CreationContext):
        self.annotation = annotation

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if isinstance(arg, self.annotation):
            return arg
        else:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

    def describe(self) -> str:
        return self.annotation.__name__

    def base_type(self) -> Any:
        return [self.annotation]
